/*
Copyright � 2014 NaturalPoint Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. */

// SimpleVRPNTest is a Windows console app that uses VRPN
// to connect to Motive and record the VRPN streaming data from
// a rigid body and an analog device, along with timing information,
// to test Motive's VRPN server.

//
// Uses VRPN v 7.33.1
//

#include <stdio.h>
#include <conio.h>
#include <chrono>
#include <thread>
#include <iostream>

// VRPN
#include "vrpn_Configure.h"             // for VRPN_CALLBACK, etc
#include "vrpn_Connection.h"			// for test server
#include "vrpn_Types.h"                 // for vrpn_float64
#include "vrpn_Tracker.h"               // for vrpn_TRACKERCB, etc
#include "vrpn_Button.h"
#include "vrpn_Analog.h"

// high resolution timing helper
#include "HiTimer.h"

// constants
const char* kButtonName = "Example1Button";		// "($MotiveButtonDeviceName)Button"
const char* kAnalogName = "Example1Analog";		// "($MotiveAnalogDeviceName)Analog"
const float kTrackerSampleRate = 100.0f;
const float kDeviceSampleRate = 200.0f;
const double kCaptureDuration = 10.0;

const bool kCSVCompatible = true;
const bool kUseLocalVRPNTestServer = false;

// frame drop/dupe detection.  Elapsed time between frame 
// should be ~= the frame period +/- a little fudge room
const double fudgeOver = 1.5;	
const double fudgeUnder = 0.5;
const int kUndefined = -99999;

// output files
FILE* gTrackerFile = nullptr;
FILE* gAnalogFile = nullptr;
FILE* gButtonFile = nullptr;

// timing information
HiTimer timer;
double gInterframeDelta = kUndefined;
int gTrackerDroppedFrames = 0;
int gTrackerDuplicateFrames = 0;
int gTrackerReceivedFrames = 0;
int gButtonDroppedFrames = 0;
int gButtonDuplicateFrames = 0;
int gButtonReceivedFrames = 0;
int gAnalogDroppedFrames = 0;
int gAnalogDuplicateFrames = 0;
int gAnalogReceivedFrames = 0;
double gTrackerPrevTime = kUndefined;
double gButtonPrevTime = kUndefined;
double gAnalogPrevTime = kUndefined;

//parameters for command line
struct ConnectionValues {
	char serverAddress[MAX_PATH];
	char localAddress[MAX_PATH];
	char portNumber[MAX_PATH];
	char objectName[MAX_PATH];
	char vrpnConnectionAddress[MAX_PATH];
	bool valid_prior_connection = false;
};

// test server devices
vrpn_Tracker_NULL* ntkr = nullptr;
vrpn_Tracker_Remote* tkr = nullptr;
vrpn_Analog_Remote* ana = nullptr;
vrpn_Connection* connection = nullptr;

extern const char* vrpn_MAGIC;
static int programSwitch(ConnectionValues& values);
static const char* VRPNVersion()
{
	unsigned int i = 0;
	for (i = 0; i < strlen(vrpn_MAGIC); i++)
		if (vrpn_MAGIC[i] >= '0' && vrpn_MAGIC[i] <= '9')
			return &vrpn_MAGIC[i];
	return vrpn_MAGIC;
}

double timestampToMilliseconds(const struct timeval t)
{
	double currentMSecs = ((double)t.tv_sec * 1000.0) + ((double)t.tv_usec * 0.001);
	return currentMSecs;
}

double getTrackerInterFrameDelta()
{
	// first frame
	if (gInterframeDelta == kUndefined)
	{
		gInterframeDelta = 0.0;
		timer.Start();
		return 0.0;
	}

	// time between frames
	timer.Stop();
	gInterframeDelta = timer.Duration();
	timer.Start();

	return gInterframeDelta * 1000.0;
}

static std::string hasDroppedOrDuped(timeval sampleTime, double prevSampleMSecs, double  sampleRate, int& drop, int& dupe, double& outElapsedSampleMSecs)
{
	if (prevSampleMSecs == kUndefined)
	{
		return "";
	}
	// record vrpn timestamps for irregularities
	double currentSampleMSecs = timestampToMilliseconds(sampleTime);
	outElapsedSampleMSecs = currentSampleMSecs - prevSampleMSecs;
	double period = (1.0 / sampleRate) * 1000.0;

	std::string hasDroppedOrDuped = "";
	if (outElapsedSampleMSecs > (period * fudgeOver))
	{
		hasDroppedOrDuped = "DROP";
		drop++;
	}
	else if (outElapsedSampleMSecs < (period * fudgeUnder))
	{
		hasDroppedOrDuped = "DUPLICATE";
		dupe++;
	}
	return hasDroppedOrDuped;
}

static void VRPN_CALLBACK handle_tracker(void* userData, const vrpn_TRACKERCB dev)
{
	double sampleRate = kTrackerSampleRate;

	// get local elapsed time
	double interframeMSecs = getTrackerInterFrameDelta();

	// get vrpn elapsed time
	double currentSampleMSecs = timestampToMilliseconds(dev.msg_time);
	double elapsedMSecs = 0.0;
	std::string err = hasDroppedOrDuped(dev.msg_time, gTrackerPrevTime, sampleRate, gTrackerDroppedFrames, gTrackerDuplicateFrames, elapsedMSecs);

	// if vrpn timstamps indicate drop/dupe, report local timing as well
	if(!err.empty())
	{
		double period = (1.0 / sampleRate) * 1000.0;
		if (interframeMSecs > (period * fudgeOver))
		{
			err += " & LOCAL DROP";
		}
		else if (interframeMSecs < (period * fudgeUnder))
		{
			err += " & LOCAL DUPLICATE";
		}
	}

	// record timestamp anomaly
	if (!err.empty())
	{
		// record frame anomaly into file
		if (kCSVCompatible)
		{
			fprintf(gTrackerFile, "%.2f,%.2f,%.2f,%s\n", currentSampleMSecs, elapsedMSecs, interframeMSecs, err.c_str());
		}
		else
		{
			fprintf(gTrackerFile, "\n[%s]\n", err.c_str());
			fprintf(gTrackerFile, "\tVRPN current sample timestamp (secs,usecs) : [%ld][%ld]\n", dev.msg_time.tv_sec, dev.msg_time.tv_usec);
			fprintf(gTrackerFile, "\tVRPN elapsed time from previous sample (msecs) : [%.2f]\n", elapsedMSecs);
		}
	}

	// record frame data
	fprintf(gTrackerFile, "%.2f,%.2f,%.2f,%.3f,%.3f,%.3f,%.3f,%.3f,%.3f,%.3f\n",
		currentSampleMSecs, elapsedMSecs, interframeMSecs, dev.pos[0], dev.pos[1], dev.pos[2], dev.quat[0], dev.quat[1], dev.quat[2], dev.quat[3]);
	
	gTrackerPrevTime = currentSampleMSecs;
	gTrackerReceivedFrames++;

}

static void VRPN_CALLBACK handle_analog(void* userData, const vrpn_ANALOGCB dev)
{
	double elapsedMSecs = 0.0;
	std::string err = hasDroppedOrDuped(dev.msg_time, gAnalogPrevTime, kDeviceSampleRate, gAnalogDroppedFrames, gAnalogDuplicateFrames, elapsedMSecs);

	double currentMSecs = timestampToMilliseconds(dev.msg_time);
	fprintf(gAnalogFile,"%.3f,%.3f,", currentMSecs, elapsedMSecs);
	for (int i = 0; i < dev.num_channel; i++)
	{
		fprintf(gAnalogFile, "%d,%.3f", i, dev.channel[i]);
	}
	fprintf(gAnalogFile, "\n");
	gAnalogPrevTime = currentMSecs;
	gAnalogReceivedFrames++;
}

static void VRPN_CALLBACK handle_button(void* userData, const vrpn_BUTTONCB dev)
{
	// note : VRPN only sends packets when values have changed.
	// In the vase of the button, timing on button off/on state change is discontinuous
	// so lets just measure and report.
	double currentSampleMSecs = timestampToMilliseconds(dev.msg_time);
	double elapsedMSecs = currentSampleMSecs - gButtonPrevTime;
	fprintf(gButtonFile, "%.3f,%.3f,%d,%d\n", currentSampleMSecs, elapsedMSecs, dev.button, dev.state);
	gButtonPrevTime = currentSampleMSecs;
	gButtonReceivedFrames++;
}



static void createOutputFiles(const char* szRootFolder, char szOption)
{
	char szFile[MAX_PATH];
	char szMode[2];
	sprintf(szFile, "%s\\%s", szRootFolder, "output_tracker.txt");
	sprintf(szMode, "%c", szOption);
	gTrackerFile = fopen(szFile, szMode);
	if (!gTrackerFile)
	{
		printf("error opening output file %s.  Exiting.", szFile);
	}
	else
	{
		fprintf(gTrackerFile, "VRPN sample timestamp(ms),VRPN sample delta(ms),Local client inter-frame delta(ms), pos x, pos y, pos z, ori x, ori y, ori z, ori w\n");
	}

	sprintf(szFile, "%s\\%s", szRootFolder, "output_analog.txt");
	gAnalogFile = fopen(szFile, szMode);
	if (!gAnalogFile)
	{
		printf("error opening output file %s.  Exiting.", szFile);
	}
	else
	{
		fprintf(gAnalogFile, "VRPN sample timestamp(ms),VRPN sample delta(ms),Analog Sensor number, Analog Value\n");
	}

	sprintf(szFile, "%s\\%s", szRootFolder, "output_button.txt");
	gButtonFile = fopen(szFile, szMode);
	if (!gButtonFile)
	{
		printf("error opening output file %s.  Exiting.", szFile);
	}
	else
	{
		fprintf(gButtonFile, "VRPN sample timestamp(ms), Elapsed Time (ms), Button number, Button State\n");
	}

}
//userPrompt function provides functionality for every attempt up to the first successful connection 
static void userPrompt(ConnectionValues& values) {
	char vrpnConnectionAddress[MAX_PATH] = "";
	char tempServerAddress[MAX_PATH] ="";
	char tempLocalAddress[MAX_PATH]="";
	char tempPort[MAX_PATH]="";
	char tempObjectName[MAX_PATH];
	printf("\nMotive IP (localhost): ");
	fgets(tempServerAddress, sizeof(tempServerAddress), stdin);
	printf("Local IP (localhost): ");
	fgets(tempLocalAddress, sizeof(tempLocalAddress), stdin);
	printf("Port Number (3883): ");
	fgets(tempPort, sizeof(tempPort), stdin);
	printf("Object Name (Rigid Body 001): ");
	fgets(tempObjectName, sizeof(tempObjectName), stdin);
	
	//comparison logic that enables for switching default vs user variables. 

	if ((strcmp(values.serverAddress, tempServerAddress) != 0) && strcmp(tempServerAddress, "\n") != 0) {
		tempServerAddress[strcspn(tempServerAddress, "\n")] = '\0';
		strcpy_s(values.serverAddress, tempServerAddress); //Motive Server
	}
	if ((strcmp(values.localAddress, tempLocalAddress) != 0) && strcmp(tempLocalAddress, "\n") != 0) {
		tempLocalAddress[strcspn(tempLocalAddress, "\n")] = '\0';
		strcpy_s(values.localAddress, tempLocalAddress); //Local NIC Address
	}
	if ((strcmp(values.portNumber, tempPort) != 0) && strcmp(tempPort, "\n") != 0) {
		tempPort[strcspn(tempPort, "\n")] = '\0';
		strcpy_s(values.portNumber , tempPort);//Port Number
	}
	if (strcmp(values.objectName, tempObjectName) != 0 && strcmp(tempObjectName, "\n") != 0) {
		tempObjectName[strcspn(tempObjectName, "\n")] = '\0';
		strcpy_s(values.objectName, tempObjectName);
		//Object Name 
	}
}


//connectedUserPrompt respects the local endpoint, and only asks for updates on other values for UI/UX clarity 
static void connectedUserPrompt(ConnectionValues& values) {
	char vrpnConnectionAddress[MAX_PATH] = "";
	char tempServerAddress[MAX_PATH] = "";
	char tempPort[MAX_PATH] = "";
	char tempObjectName[MAX_PATH];
	printf("\nMotive IP (localhost): ");
	fgets(tempServerAddress, sizeof(tempServerAddress), stdin);
	printf("Port Number (3883): ");
	fgets(tempPort, sizeof(tempPort), stdin);
	printf("Object Name (Rigid Body 001): ");
	fgets(tempObjectName, sizeof(tempObjectName), stdin);

	//comparison for the three relevant values
	if ((strcmp(values.serverAddress, tempServerAddress) != 0) && strcmp(tempServerAddress, "\n") != 0) {
		tempServerAddress[strcspn(tempServerAddress, "\n")] = '\0';
		strcpy_s(values.serverAddress, tempServerAddress); 
	}
	if ((strcmp(values.portNumber, tempPort) != 0) && strcmp(tempPort, "\n") != 0) {
		tempPort[strcspn(tempPort, "\n")] = '\0';
		strcpy_s(values.portNumber, tempPort);
	}
	if (strcmp(values.objectName, tempObjectName) != 0 && strcmp(tempObjectName, "\n") != 0) {
		tempObjectName[strcspn(tempObjectName, "\n")] = '\0';
		strcpy_s(values.objectName, tempObjectName);
	}
}

static void createConnection(ConnectionValues& values) {
		sprintf(values.vrpnConnectionAddress, "%s:%s", values.serverAddress, values.portNumber);
		vrpn_Connection* connection = vrpn_get_connection_by_name(values.vrpnConnectionAddress, nullptr, nullptr, nullptr, nullptr, values.localAddress, false);
		printf("Invalid Connection. Please Enter Valid IP Address Information. Press Enter to Continue.");
		strcpy(values.serverAddress, "localhost");
		strcpy(values.objectName, "Rigid Body 001");
		strcpy(values.portNumber, "3883");
		while (getchar() != '\n');
		printf("\n\n");
		if (values.valid_prior_connection == false) {
			strcpy(values.localAddress, "localhost");
			userPrompt(values);
		}
		else {
			connectedUserPrompt(values);
		}
		sprintf(values.vrpnConnectionAddress, "%s:%s", values.serverAddress, values.portNumber);
		return;
	}

static void runProg(ConnectionValues& values, char origOption) {
	printf("\n\nMotive IP : %s\nLocal IP : %s\nMotive Port Number: %s\nMotive Rigid Body: %s\n\n\n", values.serverAddress, values.localAddress , values.portNumber, values.objectName);
	// create output files for writing received stream into
	char szFolder[MAX_PATH];
	GetCurrentDirectory(MAX_PATH, szFolder);
	createOutputFiles(szFolder, origOption);
	printf("Output folder : %s\n", szFolder);

	//create the address that VRPN connection will listen to. 
	sprintf(values.vrpnConnectionAddress, "%s:%s", values.serverAddress, values.portNumber);
	vrpn_Connection* connection = vrpn_get_connection_by_name(values.vrpnConnectionAddress, nullptr, nullptr, nullptr, nullptr, values.localAddress, false);
	auto connect_status = connection->doing_okay();
	if (connect_status == vrpn_TRUE && values.valid_prior_connection == false) {
		values.valid_prior_connection = true;
	}
	while (connect_status != vrpn_TRUE) {
		createConnection(values);
		connection = vrpn_get_connection_by_name(values.vrpnConnectionAddress, nullptr, nullptr, nullptr, nullptr, values.localAddress, false);
		connect_status = connection->doing_okay();
		printf("\n\nMotive IP : %s\nLocal IP : %s\nMotive Port Number: %s\nMotive Rigid Body: %s\n\n\n", values.serverAddress, values.localAddress, values.portNumber, values.objectName);
		printf("Output folder : %s\n", szFolder);
	}
	
	// fire it up
	int retCode = 0;
	// Create a tracker (aka RigidBody)
	char szTrackerName[MAX_PATH];
 	sprintf(szTrackerName, "%s@%s:%s", values.objectName, values.serverAddress, values.portNumber);
	printf("Connecting to Motive RigidBody object %s...\n", szTrackerName);
	vrpn_Tracker_Remote* vrpnTracker = new vrpn_Tracker_Remote(szTrackerName);
	if (vrpnTracker)
	{
		vrpnTracker->set_update_rate(kTrackerSampleRate);
		retCode = vrpnTracker->register_change_handler(NULL, handle_tracker);
		//Conditionals for debugging purposes for tracker
		/*
		* if (retCode != 0)
			printf(" => Tracker Object Creation failed :  %d\n", retCode);
		else
			printf(" => Tracker Object Creation succeeded.\n");
		*/
	}
	else
	{
		printf("Create Tracker Error - skipping.\n");
	}

	// Create an 'Axis' (aka 'analog')
	char szAnalogName[MAX_PATH] = "";
	sprintf(szAnalogName, "%s@%s", kAnalogName, values.serverAddress);
	printf("Connecting to Motive Peripheral Analog Device %s...\n", szAnalogName);
	vrpn_Analog_Remote* vrpnAnalog = new vrpn_Analog_Remote(szAnalogName);
	if (vrpnAnalog)
	{
		vrpnAnalog->shutup = true;
		retCode = vrpnAnalog->register_change_handler(NULL, handle_analog);
		//Conditionals for debugging purposes for Analog
		/*
		* if (retCode != 0)
			printf(" => Analog Object Creation failed :  %d\n", retCode);
		else
			printf(" => Analog Object Creation succeeded.\n");
		*/
	}
	else
	{
		printf("Create Analog Error - skipping.\n");
	}

	// Create a button
	char szButtonName[MAX_PATH] = "";
	sprintf(szButtonName, "%s@%s", kButtonName, values.serverAddress);
	printf("Connecting to Motive Peripheral Button Device %s...\n", szButtonName);
	vrpn_Button_Remote* vrpnButton = new vrpn_Button_Remote(szButtonName);
	if (vrpnButton)
	{
		vrpnButton->shutup = true;
		retCode = vrpnButton->register_change_handler(NULL, handle_button);
		//Conditionals for debugging purposes for Button
		/*
		* 	if (retCode != 0)
			printf(" => Button Object Creation failed :  %d\n", retCode);
		else
			printf(" => Button Object Creation succeeded.\n");
		*/
	}
	else
	{
		printf("Create Button Error - skipping.\n");
	}

	// Capture data to output files for 10.0 seconds
	printf("\nCollecting %.2f seconds of data...\n", kCaptureDuration);
	HiTimer tm;
	tm.Start();
	while (tm.Elapsed() < kCaptureDuration)
	{
		if (vrpnTracker)
			vrpnTracker->mainloop();
		if (vrpnAnalog)
			vrpnAnalog->mainloop();
		if (vrpnButton)
			vrpnButton->mainloop();

		vrpn_SleepMsecs(0);
	}

	// cleanup
	if (vrpnTracker)
	{
		vrpnTracker->unregister_change_handler(0, handle_tracker);
		delete vrpnTracker;
	}
	if (vrpnAnalog)
	{
		vrpnAnalog->unregister_change_handler(0, handle_analog);
		delete vrpnAnalog;
	}
	if (vrpnButton)
	{
		vrpnButton->unregister_change_handler(0, handle_button);
		delete vrpnButton;
	}

	// Close up timing files
	if (gTrackerFile)
	{
		fprintf(gTrackerFile, "\nTracker Dropped Frames: %d,Duplicate Frames: %d, Received Frames: %d", gTrackerDroppedFrames, gTrackerDuplicateFrames, gTrackerReceivedFrames);
		fflush(gTrackerFile);
		fclose(gTrackerFile);
	}

	if (gAnalogFile)
	{
		fprintf(gAnalogFile, "\nAnalog Dropped Frames: %d,Duplicate Frames: %d, Received Frames: %d", gAnalogDroppedFrames, gAnalogDuplicateFrames, gAnalogReceivedFrames);
		fflush(gAnalogFile);
		fclose(gAnalogFile);
	}

	if (gButtonFile)
	{
		fprintf(gButtonFile, "\nButton Dropped Frames: %d,Duplicate Frames: %d, Received Frames: %d", gButtonDroppedFrames, gButtonDuplicateFrames, gButtonReceivedFrames);
		fflush(gButtonFile);
		fclose(gButtonFile);
	}
	//obtain the percentage 
	printf("\nCollecting Complete. \nTiming data files written to %s\n", szFolder);
	printf("Total Sample Tracker Dropped Frames: %d\n", gTrackerDroppedFrames);
	printf("Total Sample Duplicate Frames: %d\n", gTrackerDuplicateFrames);
	printf("Total Sample Received Frames: %d\n", gTrackerReceivedFrames);
	printf("\nPress r to run again with the same values, p to change the values and run again, c to clear the file and input new prompts, or any other character to quit: ");
	gTrackerDroppedFrames = 0;
	gTrackerDuplicateFrames = 0;
	gTrackerReceivedFrames = 0;
	gButtonDroppedFrames = 0;
	gButtonDuplicateFrames = 0;
	gButtonReceivedFrames = 0;
	gAnalogDroppedFrames = 0;
	gAnalogDuplicateFrames = 0;
	gAnalogReceivedFrames = 0;
	
}

//the logic that enables the program to run multiple times.
static int programSwitch(ConnectionValues& values) {
	char key;
	std::cin >> key;

	if (key != 'r' && key != 'p' && key != 'c') {
		printf("Program Complete. Quitting. ");
		return 0;
	}

	//switching
	switch (key) {
		//new user prompt
		case 'p':
		{
			while (getchar() != '\n');
			char switchOption = 'a';
			char selectedChoice = 'p';
			strcpy(values.serverAddress, "localhost");
			strcpy(values.objectName, "Rigid Body 001");
			strcpy(values.portNumber, "3883");
			connectedUserPrompt(values);
			//updates connection
			runProg(values, switchOption);
			break;
		}
		case 'r':
		//rerun without changes
		{
			printf("\n");
			char switchOption = 'a';
			char selectedChoice = 'r';
			runProg(values, switchOption);
			break;
		}
		case 'c':
		//clears the file, runs again
		{
			while (getchar() != '\n');
			char selectedChoice = 'c';
			char switchOption = 'w';
			strcpy(values.localAddress, "localhost");
			strcpy(values.serverAddress, "localhost");
			strcpy(values.objectName, "Rigid Body 001");
			strcpy(values.portNumber, "3883");
			values.valid_prior_connection = false;
			userPrompt(values);
			runProg(values, switchOption);
			break;
		}
		default:
			return 0;
			break;
		}
	//returns 1 to add in default to continue
	return 1;
	}

// SimpleVRPNTest.exe MotiveIPAddress LocalNICIPAddress
// if no addresses specified, 'localhost' is assumed
//Usage Command for Program: exe arg1 (server address) arg2 (local address) arg3 (objectName).
int main(int argc, char* argv[])
{
	printf("Optitrack VRPN Client Test.\n");
	char origOption = 'w';
	char selectedChoice = 'r';
	const char* vrpnver = VRPNVersion();
	if (vrpnver)
	{
		printf("VRPN Version : %s\n\n", VRPNVersion());
	}
	// get Motive address - default to localhost unsless cmd line args present
	//Note: For Command Line Automated Testing, all four values must be defined in the following order: 
	//ServerAddress  LocalAddress  Port  Object Name
	//default values
	ConnectionValues values;
	strcpy(values.localAddress, "localhost");
	strcpy(values.serverAddress, "localhost");
	strcpy(values.objectName, "Rigid Body 001");
	strcpy(values.portNumber, "3883");
	
	//conditionals for command line run once
	if (argc > 1) {
		if (argc >= 2) {
			strcpy_s(values.serverAddress, argv[1]);
		}
		if (argc >=3)
		{
			strcpy_s(values.localAddress, argv[2]);
		}
		if (argc >=4){
			strcpy_s(values.portNumber, argv[3]);
		}
		if (argc >=5)
		{
			strcpy_s(values.objectName, argv[4]);
		}
		runProg(values, origOption);
		return 0;
	}
	//for running multiple times in application
	else {
		//calls the userPrompt function the first time
		userPrompt(values);
		runProg(values, origOption);
		int result = 1;
		while (result != 0) {
			result = programSwitch(values);
		}
	}
	return 0;
}


